/*
    Copyright (c) 2006 Volker Krause <vkrause@kde.org>

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Library General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
    License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
*/
#ifndef CONTENT_TEST_H
#define CONTENT_TEST_H

#include <QObject>

class ContentTest : public QObject
{
    Q_OBJECT
private Q_SLOTS:
    void testGetHeaderInstance();
    void testHeaderAddRemove();
    void testHeaderAppend();
    void testImplicitMultipartGeneration();
    void testExplicitMultipartGeneration();
    void testSetContent();
    void testEncodedContent();
    void testDecodedContent();
    void testMultipartMixed();
    void testMultipleHeaderExtraction();
    /**
      Tests that a message with uuencoded content
      is parsed correctly and if a corresponding
      MIME structure is created.
    */
    void testParsingUuencoded();
    // TODO: grab samples from http://www.yenc.org/develop.htm and make a Yenc test
    void testParent();
    void testFreezing();
    void testContentTypeMimetype_data();
    void testContentTypeMimetype();
};

#endif
